/*
 * File    : comm.c
 * Purpose : define functions for parallel communication
 */

#include <avr/io.h>

#include "types.h"
#include "led.h"

/*
 * data- and control port definitions
 */
#define CTRL_PORT  PORTA
#define CTRL_PIN   PINA
#define CTRL_DDR   DDRA

#define SLEEP    PA5
#define SND      PA4
#define RD       PA3
#define WR       PA2
#define TXE      PINA1
#define RXF      PINA0

#define SLEEP_PIN 5
#define SND_PIN   4
#define RD_PIN    3
#define WR_PIN    2
#define TXE_PIN   1
#define RXF_PIN   0

#define DATA_PORT PORTB
#define DATA_PIN  PINB
#define DATA_DDR  DDRB

/* control usage of leds */
bool comm_use_red_led   = 1;
bool comm_use_green_led = 1;

/*
 * return true if RX buffer is empty
 */
int
comm_has_rx ( void )
{
    if ( bit_is_set( CTRL_PIN, RXF_PIN ) ) 
        return 0;
    else
        return 1;
}

/*
 * receive a byte from parallel interface
 */
char
comm_getc ( void )
{
    char  b;
    
    /* wait until a byte is in RX-FIFO */
    while ( ! comm_has_rx() )
        ;

    /* read the byte */
    outb( DATA_DDR, 0x00 );
    cbi( CTRL_PORT, RD );

    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );

    b = inb( DATA_PIN );

    sbi( CTRL_PORT, RD );

    return b;
}

/*
 * receive whole line from parallel interface
 */
int
comm_gets ( char * buf, int size )
{
    uchar  i = 0;
    char   b;

    if ( comm_use_green_led )
        led_on( LED_GREEN );
    
    while ( i < size-1 )
    {
        b = comm_getc();
            
        if (( b == '\n' ) || ( b == '\r' ))
            break;
        
        buf[i++] = b;
    }

    buf[i] = '\0';

    if ( comm_use_green_led )
        led_off_d( LED_GREEN );
    
    return i;
}

/*
 * transmit byte via parallel interface
 */
int
comm_putc ( char b )
{
    /* wait until FIFO is not full */
    while ( bit_is_set( CTRL_PORT, TXE_PIN ) )
        ;

    /* read the byte */
    outb( DATA_DDR, 0xff );
    outb( DATA_PORT, b );
    
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    cbi( CTRL_PORT, WR_PIN );
    
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    sbi( CTRL_PORT, WR_PIN );
    
    outb( DATA_DDR, 0x00 );

    return 1;
}

/*
 * transmit line via parallel interface
 */
int
comm_puts ( char * buf )
{
    if ( comm_use_red_led )
        led_on( LED_RED );
    
    while ( *buf )
        comm_putc( *buf++ );

    if ( comm_use_red_led )
        led_off_d( LED_RED );

    return 0;
}

/*
 * initialise parallel communication
 */
void
comm_init ( void )
{
    /* set RXF and TXE as input */
    cbi( CTRL_DDR, RXF );
    cbi( CTRL_DDR, TXE );
    cbi( CTRL_DDR, SLEEP );

    /* and WR and RD as output (HIGH) */
    sbi( CTRL_DDR, WR );
    sbi( CTRL_DDR, RD );
    sbi( CTRL_DDR, SND );

    sbi( CTRL_PORT, WR_PIN );
    sbi( CTRL_PORT, RD_PIN );
    cbi( CTRL_PORT, SND_PIN );
}

