/*
 * File    : lcd.c
 * Purpose : define functions to control LCD
 * Author  : Ronald Kriemann
 * Date    : 26.11.2003
 */

#include <avr/io.h>
#include <string.h>

#include "delay.h"

/* lcd port */
#define LCD_DDR    DDRC
#define LCD_PORT   PORTC
#define PIN_RS     4
#define PORT_EN    PC5

/* local definitions */
#define LCD_CLEAR        0
#define LCD_HOME         1
#define LCD_ENTRY        2
#define LCD_DISP_CTRL    3
#define LCD_CUROR_SHIFT  4
#define LCD_FUNC_SET     5
#define LCD_SET_CGRAM    6
#define LCD_SET_DDRAM    7

#define LCD_SHIFT        0
#define LCD_INC          1

#define LCD_BLINK        0
#define LCD_CURSOR_ON    1
#define LCD_DISP_ON      2

#define LCD_8BIT         4
#define LCD_2LINES       3

#define LCD_ADR_LINE1  0x00
#define LCD_ADR_LINE2  0x40
#define LCD_ADR_LINE3  0x14
#define LCD_ADR_LINE4  0x54

/*
 * tell LCD to take over data
 */
void
lcd_enable ( void )
{
    sbi( LCD_PORT, PORT_EN );

    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    __asm__ __volatile__( "nop" );
    
    cbi( LCD_PORT, PORT_EN );
}

/*
 * write command (rs=0) or data (rs=1) to LCD
 */
void
lcd_write ( uchar data, uchar rs )
{
    if ( rs )
    {
        LCD_PORT = ((data >> 4) & 0x0F) | (1 << PIN_RS);
        lcd_enable();

        LCD_PORT = (data & 0x0F) | (1 << PIN_RS);
        lcd_enable();
    }
    else
    {
        LCD_PORT = ((data >> 4) & 0x0F);
        lcd_enable();

        LCD_PORT = (data & 0x0F);
        lcd_enable();
    }

    delay_us( 60 );
}

/*
 * put a single character on LCD
 */
void
lcd_putc ( uchar c )
{
    lcd_write( c, 1 );
}

/*
 * put string on LCD
 */
void
lcd_puts ( const uchar * s )
{
    uchar i = 0;
    uchar n = strlen( s );

    while ( i < n )
        lcd_putc( s[i++] );
}

/*
 * set cursor at given pos
 */
void
lcd_gotoxy ( uchar x, uchar y )
{
    switch ( y )
    {
    case 0 : lcd_write( (1 << LCD_SET_DDRAM) + LCD_ADR_LINE1 + x, 0 ); break;
    case 1 : lcd_write( (1 << LCD_SET_DDRAM) + LCD_ADR_LINE2 + x, 0 ); break;
    case 2 : lcd_write( (1 << LCD_SET_DDRAM) + LCD_ADR_LINE3 + x, 0 ); break;
    case 3 : lcd_write( (1 << LCD_SET_DDRAM) + LCD_ADR_LINE4 + x, 0 ); break;
    }
}

/*
 * clear LCD
 */
void
lcd_clear ( void )
{
    lcd_write( (1 << LCD_CLEAR), 0 );
    delay_ms( 2 );
}

/* sets cursor on and off */
void
lcd_cursor_on ( void )
{
    lcd_write( (1 << LCD_DISP_CTRL) |
               (1 << LCD_DISP_ON)   |
               (1 << LCD_CURSOR_ON ), 0 );
}

void
lcd_cursor_off ( void )
{
    lcd_write( (1 << LCD_DISP_CTRL) |
               (1 << LCD_DISP_ON), 0 );
}

void
lcd_cursor_blink ( void )
{
    lcd_write( (1 << LCD_DISP_CTRL)  |
               (1 << LCD_DISP_ON)    |
               (1 << LCD_BLINK), 0 );
}

/*
 * set n'th programmable character
 */
void
lcd_set_char ( uchar n, uchar * data )
{
    uchar  i;
    
    lcd_write( (1 << LCD_SET_CGRAM) | (8 * n), 0 );

    for ( i = 0; i < 8; i++ )
        lcd_write( data[i], 1 );
}

/*
 * create horizontal bars
 */
void
lcd_hbars ( void )
{
    uchar  cdat[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
    uchar  i, j;

    for ( j = 1; j < 5; j++ )
    {
        /* 10000 ... 11110 */
        for ( i = 0; i < 8; i++ )
            cdat[i] = cdat[i] | (1 << (5-j));
        lcd_set_char( j, cdat );
    }
}

/*
 * create vertikal bars
 */
void
lcd_vbars ( void )
{
    uchar  cdat[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
    uchar  i;

    for ( i = 0; i < 7; i++ )
    {
        /* line i */
        cdat[7-i] = 0xff;
        lcd_set_char( i+1, cdat );
    }
}

/*
 * init LCD
 */
void
lcd_init ( void )
{
    /* init pins for operation */
    LCD_DDR = 0x3f;

    /* wait after powerup */
    delay_ms( 50 );

    /* 3 times send 0x3 */
    LCD_PORT = 0x03;
    lcd_enable();
    delay_ms( 5 );

    LCD_PORT = 0x03;
    lcd_enable();
    delay_ms( 5 );

    LCD_PORT = 0x03;
    lcd_enable();
    delay_ms( 5 );

    /* send 0x2 to enable 4-bit mode */
    LCD_PORT = 0x02;
    lcd_enable();
    delay_ms( 5 );

    /* 2 lines, 4 bit */
    lcd_write( (1 << LCD_FUNC_SET) |
               (1 << LCD_2LINES ), 0 );

    /* display on, cursor on */
    lcd_write( (1 << LCD_DISP_CTRL) |
               (1 << LCD_DISP_ON)   |
               (1 << LCD_CURSOR_ON), 0 );

    /* inc ddram after write */
    lcd_write( (1 << LCD_ENTRY) |
               (1 << LCD_INC), 0 );
}

